#include "boolean.h"
#include "sheet.h"
#include "plotter.h"

static struct {
	position pos;
	bool pen_down;
	int pen;		/* n de stylo courant */
	int direction;	/* direction d'criture */
	int char_size;	/* taille caractre: 0  63 */
	bool graphics;
	position origin;
	int line_type;	/* espacement des pointills: 0  15 */
} plotter;

void inc_x()
{
	if (plotter.pos.x < XMAX)
		plotter.pos.x++;
}

void dec_x()
{
	if (plotter.pos.x > 0)
		plotter.pos.x--;
}

void inc_y()
{
	if (roll_sheet(plotter.pos.y+1))
		plotter.pos.y++;
}

void dec_y()
{
	if (plotter.pos.y > 0)
		plotter.pos.y--;
}

void change_color(int pen)
{
	if (pen>=0 && pen<=3)
		plotter.pen=pen;
}

void set_dir(int dir)
{
	if (dir>=0 && dir<=3)
		plotter.direction=dir;
}

void home() { plotter.pos=plotter.origin; }
void set_origin() { plotter.origin=plotter.pos; }
void set_graphic_mode() { plotter.graphics=true; set_origin(); }
void set_text_mode() { plotter.graphics=false; }

void set_line_type(int line_type)
{
	if (line_type>=0 && line_type<=15)
		plotter.line_type=line_type;
}

void set_char_size(int char_size)
{
	if (char_size>=0 && char_size<=63)
		plotter.char_size=char_size;
}

void plot(position pos, int colour)
{
	if (roll_sheet(pos.y))
		mark_point(pos.x,colour+1);
}

void line_plot(position pos,int colour)
{
	static int cnt=0;
	static bool draw=false;
	if (plotter.line_type==0) draw=true;
	else {
		cnt++;
		if (cnt>=plotter.line_type) {
			draw = !draw;
			cnt=0;
		}
	}
	if (draw)
		plot(pos,colour);
}

void pen_down()
{
	plotter.pen_down=true;
	plot(plotter.pos,plotter.pen);
}

void pen_down_no_mark()
{
	plotter.pen_down=true;
}

void pen_up() {	plotter.pen_down=false; }
void rotate_color() { plotter.pen=(plotter.pen+1)%NBCOLORS; }
int get_dir() {	return plotter.direction; }
int get_char_size() { return plotter.char_size; }
position get_pen_pos() { return plotter.pos; }
position get_origin() { return plotter.origin; }
void set_pen_pos(position newpos) { plotter.pos=newpos; }
bool is_graphic_mode() { return plotter.graphics; }

void pen_step(int absolute_dir)
{
	switch (absolute_dir) {
	case RIGHT:		inc_x(); break;
	case UPRIGHT:	inc_x();
	case UP:		dec_y(); break;
	case UPLEFT:	dec_y();
	case LEFT:		dec_x(); break;
	case DOWNLEFT:	dec_x();
	case DOWN:		inc_y(); break;
	case DOWNRIGHT: inc_y(); inc_x();
	}
	if (plotter.pen_down)
		plot(plotter.pos,plotter.pen);
}

void move_pen(int relative_dir,int length)
{
	int i;
	int absolute_dir = (8-2*get_dir()+relative_dir)%8;
	for (i=0;i<length;i++)
		pen_step(absolute_dir);
}

void draw_to(int xdest, int ydest)
{
	int i;
	int dx=xdest-plotter.pos.x;
	int dy=ydest-plotter.pos.y;
	line_plot(plotter.pos,plotter.pen);
	if (dx==0 && dy==0) return;
	if ( abs(dx) < abs(dy) ) {
		double frac=0.0;
		double step=((double)abs(dx))/((double)abs(dy));
		for (i=0;i<abs(dy);i++) {
			if (dy>0) inc_y();
			else dec_y();
			frac+=step;
			if (frac>0.5) {
				frac-=1.0;
				if (dx>0) inc_x();
				else dec_x();
			}
			line_plot(plotter.pos,plotter.pen);
		}
	} else {
		double frac=0.0;
		double step=((double)abs(dy))/((double)abs(dx));
		for (i=0;i<abs(dx);i++) {
			if (dx>0) inc_x();
			else dec_x();
			frac+=step;
			if (frac>0.5) {
				frac-=1.0;
				if (dy>0) inc_y();
				else dec_y();
			}
			line_plot(plotter.pos,plotter.pen);
		}
	}
}

void init_plotter()
{
	int i;
	plotter.char_size=0;	/* suivant la position du switch arrire: 0 pour 80 colonnes, 1 pour 40 colonnes */
	plotter.pos.y = -1;
	for (i=0;i<8;i++)
		pen_step(DOWN);
}


